@@ -25,6 +25,7 @@
 #include "video_renderer.h"
 
 #define SECOND_IN_NSECS 1000000000UL
+#define SECOND_IN_MICROSECS 1000000
 #ifdef X_DISPLAY_FIX
 #include <gst/video/navigation.h>
 #include "x_display_fix.h"
@@ -45,6 +46,31 @@ static bool use_x11 = false;
 #endif
 static bool logger_debug = false;
 static bool video_terminate = false;
+static gint64 hls_requested_start_position = 0;
+static gint64 hls_seek_start = 0;
+static gint64 hls_seek_end = 0;
+static gint64 hls_duration;
+static gboolean hls_seek_enabled;
+static gboolean hls_playing;
+static gboolean hls_buffer_empty;
+static gboolean hls_buffer_full;
+
+
+typedef enum {
+  //GST_PLAY_FLAG_VIDEO         = (1 << 0),
+  //GST_PLAY_FLAG_AUDIO         = (1 << 1),
+  //GST_PLAY_FLAG_TEXT          = (1 << 2),
+  //GST_PLAY_FLAG_VIS           = (1 << 3),
+  //GST_PLAY_FLAG_SOFT_VOLUME   = (1 << 4),
+  //GST_PLAY_FLAG_NATIVE_AUDIO  = (1 << 5),
+  //GST_PLAY_FLAG_NATIVE_VIDEO  = (1 << 6),
+  GST_PLAY_FLAG_DOWNLOAD      = (1 << 7),
+  GST_PLAY_FLAG_BUFFERING     = (1 << 8),
+  //GST_PLAY_FLAG_DEINTERLACE   = (1 << 9),
+  //GST_PLAY_FLAG_SOFT_COLORBALANCE = (1 << 10),
+  //GST_PLAY_FLAG_FORCE_FILTERS = (1 << 11),
+  //GST_PLAY_FLAG_FORCE_SW_DECODERS = (1 << 12),
+} GstPlayFlags;
 
 #define NCODECS  2   /* renderers for h264 and h265 */
 
@@ -52,7 +78,7 @@ struct video_renderer_s {
     GstElement *appsrc, *pipeline;
     GstBus *bus;
     const char *codec;
-    bool autovideo, state_pending;
+    bool autovideo;
     int id;
     gboolean terminate;
     gint64 duration;
@@ -128,14 +154,14 @@ static void append_videoflip (GString *launch, const videoflip_t *flip, const vi
     }
 }
 
-/* apple uses colorimetry=1:3:5:1                                *
+/* apple uses colorimetry that is detected as  1:3:7:1           * //previously 1:3:5:1 was seen
  * (not recognized by v4l2 plugin in Gstreamer  < 1.20.4)        *
  * See .../gst-libs/gst/video/video-color.h in gst-plugins-base  *
  * range = 1   -> GST_VIDEO_COLOR_RANGE_0_255      ("full RGB")  * 
  * matrix = 3  -> GST_VIDEO_COLOR_MATRIX_BT709                   *
- * transfer = 5 -> GST_VIDEO_TRANSFER_BT709                      *
+ * transfer = 7 -> GST_VIDEO_TRANSFER_SRGB                       * // previously GST_VIDEO_TRANSFER_BT709
  * primaries = 1 -> GST_VIDEO_COLOR_PRIMARIES_BT709              *
- * closest used by  GStreamer < 1.20.4 is BT709, 2:3:5:1 with    *                            *
+ * closest used by  GStreamer < 1.20.4 is BT709, 2:3:5:1 with    * // now use sRGB = 1:1:7:1    
  * range = 2 -> GST_VIDEO_COLOR_RANGE_16_235 ("limited RGB")     */  
 
 static const char h264_caps[]="video/x-h264,stream-format=(string)byte-stream,alignment=(string)au";
@@ -156,7 +182,7 @@ GstElement *make_video_sink(const char *videosink, const char *videosink_options
         return NULL;
     }
 
-    /* process the video_sink_optons */
+    /* process the video_sink_options */
     size_t len = strlen(videosink_options);
     if (!len) {
         return video_sink;
@@ -193,7 +219,7 @@ GstElement *make_video_sink(const char *videosink, const char *videosink_options
 
 void  video_renderer_init(logger_t *render_logger, const char *server_name, videoflip_t videoflip[2], const char *parser,
                           const char *decoder, const char *converter, const char *videosink, const char *videosink_options, 
-                          bool initial_fullscreen, bool video_sync, bool h265_support, const char *uri) {
+                          bool initial_fullscreen, bool video_sync, bool h265_support, guint playbin_version, const char *uri) {
     GError *error = NULL;
     GstCaps *caps = NULL;
     hls_video = (uri != NULL);
@@ -203,6 +229,14 @@ void  video_renderer_init(logger_t *render_logger, const char *server_name, vide
     logger = render_logger;
     logger_debug = (logger_get_level(logger) >= LOGGER_DEBUG);
     video_terminate = false;
+    hls_seek_enabled = FALSE;
+    hls_playing = FALSE;
+    hls_seek_start = -1;
+    hls_seek_end = -1;
+    hls_duration = -1;
+    hls_buffer_empty = TRUE;
+    hls_buffer_empty = FALSE;
+    
 
     /* this call to g_set_application_name makes server_name appear in the  X11 display window title bar, */
     /* (instead of the program name uxplay taken from (argv[0]). It is only set one time. */
@@ -227,18 +261,28 @@ void  video_renderer_init(logger_t *render_logger, const char *server_name, vide
         renderer_type[i] = (video_renderer_t *) calloc(1, sizeof(video_renderer_t));
         g_assert(renderer_type[i]);
         renderer_type[i]->autovideo = auto_videosink;
-	renderer_type[i]->id = i;
-	renderer_type[i]->bus = NULL;
+        renderer_type[i]->id = i;
+        renderer_type[i]->bus = NULL;
         if (hls_video) {
-            /* use playbin3 to play HLS video: replace "playbin3" by "playbin" to use playbin2 */ 
-            renderer_type[i]->pipeline = gst_element_factory_make("playbin3", "hls-playbin3");
+            /* use playbin3 to play HLS video: replace "playbin3" by "playbin" to use playbin2 */
+            switch (playbin_version)  {
+            case 2:
+                renderer_type[i]->pipeline = gst_element_factory_make("playbin", "hls-playbin2");
+                break;
+            case 3:
+                renderer_type[i]->pipeline = gst_element_factory_make("playbin3", "hls-playbin3");
+                break;
+            default:
+                logger_log(logger, LOGGER_ERR, "video_renderer_init: invalid playbin version %u", playbin_version);
+                g_assert(0);
+            }
+            logger_log(logger, LOGGER_INFO, "Will use GStreamer playbin version %u to play HLS streamed video", playbin_version);	    
             g_assert(renderer_type[i]->pipeline);
             renderer_type[i]->appsrc = NULL;
 	    renderer_type[i]->codec = hls;
-            /* if we are not using autovideosink, build a videossink based on the stricng "videosink" */
+            /* if we are not using autovideosink, build a videossink based on the string "videosink" */
             if(strcmp(videosink, "autovideosink")) {
-	      GstElement *playbin_videosink = make_video_sink(videosink, videosink_options);
-	      
+                GstElement *playbin_videosink = make_video_sink(videosink, videosink_options);  
                 if (!playbin_videosink) {
                     logger_log(logger, LOGGER_ERR, "video_renderer_init: failed to create playbin_videosink");
                 } else {
@@ -246,8 +290,12 @@ void  video_renderer_init(logger_t *render_logger, const char *server_name, vide
                     g_object_set(G_OBJECT (renderer_type[i]->pipeline), "video-sink", playbin_videosink, NULL);
                 }
             }
-
-            g_object_set (G_OBJECT (renderer_type[i]->pipeline), "uri", uri, NULL);
+            gint flags;
+            g_object_get(renderer_type[i]->pipeline, "flags", &flags, NULL);
+            flags |= GST_PLAY_FLAG_DOWNLOAD;
+	    flags |= GST_PLAY_FLAG_BUFFERING;    // set by default in playbin3, but not in playbin2; is it needed?
+            g_object_set(renderer_type[i]->pipeline, "flags", flags, NULL);
+	    g_object_set (G_OBJECT (renderer_type[i]->pipeline), "uri", uri, NULL);
         } else {
             switch (i) {
             case 0:
@@ -302,7 +350,13 @@ void  video_renderer_init(logger_t *render_logger, const char *server_name, vide
             logger_log(logger, LOGGER_DEBUG, "GStreamer video pipeline %d:\n\"%s\"", i + 1, launch->str);
             renderer_type[i]->pipeline = gst_parse_launch(launch->str, &error);
             if (error) {
-                g_error ("get_parse_launch error (video) :\n %s\n",error->message);
+                logger_log(logger, LOGGER_ERR, "GStreamer gst_parse_launch failed to create video pipeline %d\n"
+                           "*** error message from gst_parse_launch was:\n%s\n"
+                           "launch string parsed was \n[%s]", i + 1, error->message, launch->str);
+		if (strstr(error->message, "no element")) {
+                    logger_log(logger, LOGGER_ERR, "This error usually means that a uxplay option was mistyped\n"
+                               "           or some requested part of GStreamer is not installed\n");
+                }
                 g_clear_error (&error);
             }
             g_assert (renderer_type[i]->pipeline);
@@ -325,11 +379,14 @@ void  video_renderer_init(logger_t *render_logger, const char *server_name, vide
         renderer_type[i]->gst_window = NULL;
         renderer_type[i]->use_x11 = false;
         X11_search_attempts = 0;
+	/* setting char *x11_display_name to NULL means the value is taken from $DISPLAY in the environment 
+         * (a uxplay option to specify a different value is possible)  */
+	char *x11_display_name = NULL;
         if (use_x11) {
             if (i == 0) {
                 renderer_type[0]->gst_window = (X11_Window_t *) calloc(1, sizeof(X11_Window_t));
                 g_assert(renderer_type[0]->gst_window);
-                get_X11_Display(renderer_type[0]->gst_window);
+                get_X11_Display(renderer_type[0]->gst_window, x11_display_name);
                 if (renderer_type[0]->gst_window->display) {
                     renderer_type[i]->use_x11 = true;
                 } else {
@@ -385,18 +442,23 @@ void video_renderer_resume() {
 }
 
 void video_renderer_start() {
+    GstState state;
+    const gchar *state_name;
     if (hls_video) {
         renderer->bus = gst_element_get_bus(renderer->pipeline);
-        gst_element_set_state (renderer->pipeline, GST_STATE_PLAYING);
+        gst_element_set_state (renderer->pipeline, GST_STATE_PAUSED);
+	gst_element_get_state(renderer->pipeline, &state, NULL, 1000 * GST_MSECOND);
+	state_name= gst_element_state_get_name(state);
+	logger_log(logger, LOGGER_DEBUG, "video renderer_start: state %s", state_name);
         return;
     } 
   /* when not hls, start both h264 and h265 pipelines; will shut down the "wrong" one when we know the codec */
     for (int i = 0; i < n_renderers; i++) {
-        gst_element_set_state (renderer_type[i]->pipeline, GST_STATE_PLAYING);
-	if (renderer_type[i]->appsrc) {
-            gst_video_pipeline_base_time = gst_element_get_base_time(renderer_type[i]->appsrc);
-        }
         renderer_type[i]->bus = gst_element_get_bus(renderer_type[i]->pipeline);
+        gst_element_set_state (renderer_type[i]->pipeline, GST_STATE_PAUSED);
+	gst_element_get_state(renderer_type[i]->pipeline, &state, NULL, 1000 * GST_MSECOND);
+	state_name= gst_element_state_get_name(state);
+	logger_log(logger, LOGGER_DEBUG, "video renderer_start: renderer %d state %s", i, state_name);
     }
     renderer = NULL;
     first_packet = true;
@@ -418,11 +480,14 @@ bool waiting_for_x11_window() {
 	    return true;    /* window still not found */
         }
     }
+    if (fullscreen) {
+         set_fullscreen(renderer->gst_window, &fullscreen);
+    }
 #endif
     return false;
 }
 
-void video_renderer_render_buffer(unsigned char* data, int *data_len, int *nal_count, uint64_t *ntp_time) {
+uint64_t video_renderer_render_buffer(unsigned char* data, int *data_len, int *nal_count, uint64_t *ntp_time) {
     GstBuffer *buffer;
     GstClockTime pts = (GstClockTime) *ntp_time; /*now in nsecs */
     //GstClockTimeDiff latency = GST_CLOCK_DIFF(gst_element_get_current_clock_time (renderer->appsrc), pts);
@@ -430,9 +495,10 @@ void video_renderer_render_buffer(unsigned char* data, int *data_len, int *nal_c
         if (pts >= gst_video_pipeline_base_time) {
             pts -= gst_video_pipeline_base_time;
         } else {
-            logger_log(logger, LOGGER_ERR, "*** invalid ntp_time < gst_video_pipeline_base_time\n%8.6f ntp_time\n%8.6f base_time",
+            // adjust timestamps to be >= gst_video_pipeline_base time
+            logger_log(logger, LOGGER_DEBUG, "*** invalid ntp_time < gst_video_pipeline_base_time\n%8.6f ntp_time\n%8.6f base_time",
                        ((double) *ntp_time) / SECOND_IN_NSECS, ((double) gst_video_pipeline_base_time) / SECOND_IN_NSECS);
-            return;
+            return  (uint64_t)  gst_video_pipeline_base_time - pts;
         }
     }
     g_assert(data_len != 0);
@@ -469,6 +535,7 @@ void video_renderer_render_buffer(unsigned char* data, int *data_len, int *nal_c
         }
 #endif
     }
+    return 0;
 }
 
 void video_renderer_flush() {
@@ -476,6 +543,7 @@ void video_renderer_flush() {
 
 void video_renderer_stop() {
     if (renderer) {
+        logger_log(logger, LOGGER_DEBUG,"video_renderer_stop");
         if (renderer->appsrc) {
             gst_app_src_end_of_stream (GST_APP_SRC(renderer->appsrc));
         }
@@ -484,15 +552,21 @@ void video_renderer_stop() {
      }
 }
 
-static void video_renderer_destroy_h26x(video_renderer_t *renderer) {
+static void video_renderer_destroy_instance(video_renderer_t *renderer) {
     if (renderer) {
+        logger_log(logger, LOGGER_DEBUG,"destroying renderer instance %p", renderer);
         GstState state;
+	GstStateChangeReturn ret;
         gst_element_get_state(renderer->pipeline, &state, NULL, 100 * GST_MSECOND);
+	logger_log(logger, LOGGER_DEBUG,"pipeline state is %s", gst_element_state_get_name(state));
         if (state != GST_STATE_NULL) {
             if (!hls_video) {
                 gst_app_src_end_of_stream (GST_APP_SRC(renderer->appsrc));
             }
-            gst_element_set_state (renderer->pipeline, GST_STATE_NULL);
+            ret = gst_element_set_state (renderer->pipeline, GST_STATE_NULL);
+	    logger_log(logger, LOGGER_DEBUG,"pipeline_state_change_return: %s",
+		       gst_element_state_change_return_get_name(ret));
+	    gst_element_get_state(renderer->pipeline, NULL, NULL, 1000 * GST_MSECOND);
         }
         gst_object_unref(renderer->bus);
 	if (renderer->appsrc) {
@@ -513,53 +587,139 @@ static void video_renderer_destroy_h26x(video_renderer_t *renderer) {
 void video_renderer_destroy() {
     for (int i = 0; i < n_renderers; i++) {
         if (renderer_type[i]) {
-            video_renderer_destroy_h26x(renderer_type[i]);
+            video_renderer_destroy_instance(renderer_type[i]);
         }
-    }  
+    }
 }
 
+static void get_stream_status_name(GstStreamStatusType type, char *name, size_t len) {
+  switch (type) {
+  case GST_STREAM_STATUS_TYPE_CREATE:
+    strncpy(name, "CREATE", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_ENTER:
+    strncpy(name, "ENTER", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_LEAVE:
+    strncpy(name, "LEAVE", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_DESTROY:
+    strncpy(name, "DESTROY", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_START:
+    strncpy(name, "START", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_PAUSE:
+    strncpy(name, "PAUSE", len);
+    return;
+  case GST_STREAM_STATUS_TYPE_STOP:
+    strncpy(name, "STOP", len);
+    return;
+  default:
+    strncpy(name, "", len);
+    return;
+  }
+}
+  
 gboolean gstreamer_pipeline_bus_callback(GstBus *bus, GstMessage *message, void *loop) {
+    GstState old_state, new_state;
+    const gchar no_state[] = "";
+    const gchar *old_state_name = no_state, *new_state_name = no_state;
+    if (GST_MESSAGE_TYPE(message) == GST_MESSAGE_STATE_CHANGED) {
+        GstState old_state, new_state;
+        gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
+        old_state_name = gst_element_state_get_name (old_state);
+        new_state_name = gst_element_state_get_name (new_state);
+    }
 
-  /* identify which pipeline sent the message */ 
+    /* identify which pipeline sent the message */ 
     int type = -1;
     for (int i = 0 ; i < n_renderers ; i ++ ) {
-        if (renderer_type[i]->bus == bus) {
+        if (renderer_type[i] && renderer_type[i]->bus == bus) {
             type = i;
             break;
         }
     }
-    g_assert(type != -1);
 
-    if (logger_debug) {
-        g_print("GStreamer %s bus message: %s %s\n", renderer_type[type]->codec, GST_MESSAGE_SRC_NAME(message), GST_MESSAGE_TYPE_NAME(message));
+    /* if the bus sending the message is not found, the renderer may already have been destroyed */
+    if (type == -1) {
+        if (logger_debug) {
+            g_print("GStreamer(UNKNOWN, now destroyed?) bus message: %s %s %s %s\n",
+                     GST_MESSAGE_SRC_NAME(message), GST_MESSAGE_TYPE_NAME(message), old_state_name, new_state_name);
+        }     
+        return TRUE;
     }
 
-    if (logger_debug && hls_video) {
-        gint64 pos;
-        gst_element_query_position (renderer_type[type]->pipeline, GST_FORMAT_TIME, &pos);
+    if (logger_debug) {
+        gchar *name = NULL;
+	GstElement *element = NULL;
+        gchar type_name[8] = { 0 };
+        if (GST_MESSAGE_TYPE(message) == GST_MESSAGE_STREAM_STATUS) {
+            GstStreamStatusType type;
+            gst_message_parse_stream_status(message, &type, &element);
+            name = gst_element_get_name(element);
+            get_stream_status_name(type, type_name, 8);
+            old_state_name = name;
+            new_state_name = type_name;
+        }
+        gint64 pos = -1;
+        if (hls_video) {
+            gst_element_query_position (renderer_type[type]->pipeline, GST_FORMAT_TIME, &pos);
+        }
         if (GST_CLOCK_TIME_IS_VALID(pos)) {
-            g_print("GStreamer bus message %s %s; position: %" GST_TIME_FORMAT "\n", GST_MESSAGE_SRC_NAME(message),
-            GST_MESSAGE_TYPE_NAME(message), GST_TIME_ARGS(pos));
+            g_print("GStreamer %s  bus message %s %s %s %s; position: %" GST_TIME_FORMAT "\n" ,renderer_type[type]->codec,
+                     GST_MESSAGE_SRC_NAME(message), GST_MESSAGE_TYPE_NAME(message), old_state_name, new_state_name, GST_TIME_ARGS(pos));
         } else {
-            g_print("GStreamer bus message %s %s; position: none\n", GST_MESSAGE_SRC_NAME(message),
-                   GST_MESSAGE_TYPE_NAME(message));
+            g_print("GStreamer %s bus message %s %s %s %s\n", renderer_type[type]->codec,
+                    GST_MESSAGE_SRC_NAME(message), GST_MESSAGE_TYPE_NAME(message), old_state_name, new_state_name);
+        }
+	if (name) {
+	  g_free(name);
+	}
+    }
+
+    /* monitor hls video position until seek to hls_start_position is achieved */
+    if (hls_video && hls_requested_start_position) {
+        if (strstr(GST_MESSAGE_SRC_NAME(message), "sink")) {	  
+            gint64 pos;
+            if (!GST_CLOCK_TIME_IS_VALID(hls_duration)) {
+                gst_element_query_duration (renderer->pipeline, GST_FORMAT_TIME, &hls_duration);
+            }
+	    gst_element_query_position (renderer_type[type]->pipeline, GST_FORMAT_TIME, &pos);
+            //g_print("HLS position %" GST_TIME_FORMAT " requested_start_position %" GST_TIME_FORMAT " duration %" GST_TIME_FORMAT " %s\n",
+            //    GST_TIME_ARGS(pos), GST_TIME_ARGS(hls_requested_start_position), GST_TIME_ARGS(hls_duration),
+            //    (hls_seek_enabled ? "seek enabled" : "seek not enabled"));
+            if (pos > hls_requested_start_position) {
+                hls_requested_start_position = 0;
+            }
+	    if ( hls_requested_start_position && pos < hls_requested_start_position  && hls_seek_enabled) {
+                g_print("***************** seek to hls_requested_start_position %" GST_TIME_FORMAT "\n", GST_TIME_ARGS(hls_requested_start_position));
+                if (gst_element_seek_simple (renderer_type[type]->pipeline, GST_FORMAT_TIME,
+                                            GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT, hls_requested_start_position)) {
+                    hls_requested_start_position = 0;
+                }
+	    }
 	}
     }
 
     switch (GST_MESSAGE_TYPE (message)) {
     case GST_MESSAGE_DURATION:
-        renderer_type[type]->duration = GST_CLOCK_TIME_NONE;
+        hls_duration = GST_CLOCK_TIME_NONE;
         break;
     case GST_MESSAGE_BUFFERING:
         if (hls_video) {
             gint percent = -1;
             gst_message_parse_buffering(message, &percent);
-	    if (percent >= 0) {
+            hls_buffer_empty = TRUE;
+	    hls_buffer_full = FALSE;
+	    if (percent > 0) {
+                hls_buffer_empty = FALSE;
                 renderer_type[type]->buffering_level = percent;
-                logger_log(logger, LOGGER_DEBUG, "Buffering :%u percent done", percent);
+                logger_log(logger, LOGGER_DEBUG, "Buffering :%d percent done", percent);
                 if (percent < 100) {
                     gst_element_set_state (renderer_type[type]->pipeline, GST_STATE_PAUSED);
                 } else {
+                    hls_buffer_full = TRUE;
                     gst_element_set_state (renderer_type[type]->pipeline, GST_STATE_PLAYING);
                 }
             }
@@ -603,14 +763,31 @@ gboolean gstreamer_pipeline_bus_callback(GstBus *bus, GstMessage *message, void
         }
         break;
     case GST_MESSAGE_STATE_CHANGED:
-        if (renderer_type[type]->state_pending && strstr(GST_MESSAGE_SRC_NAME(message), "pipeline")) {
-            GstState state;
-            gst_element_get_state(renderer_type[type]->pipeline, &state, NULL, 100 * GST_MSECOND);
-	    if (state == GST_STATE_NULL) {
-                gst_element_set_state(renderer_type[type]->pipeline, GST_STATE_PLAYING);
-	    } else if (state == GST_STATE_PLAYING) {
-                renderer_type[type]->state_pending = false;
+        if (hls_video && logger_debug && strstr(GST_MESSAGE_SRC_NAME(message), "hls-playbin")) {
+            GstState old_state, new_state;
+            gst_message_parse_state_changed (message, &old_state, &new_state, NULL);
+            g_print ("****** hls_playbin: Element %s changed state from %s to %s.\n", GST_OBJECT_NAME (message->src),
+                     gst_element_state_get_name (old_state),
+                     gst_element_state_get_name (new_state));
+            if (new_state != GST_STATE_PLAYING) {
+                break;
+                hls_playing = FALSE;
             }
+            hls_playing = TRUE;
+            GstQuery *query;
+            query = gst_query_new_seeking(GST_FORMAT_TIME);
+                if (gst_element_query(renderer->pipeline, query)) {
+	        gst_query_parse_seeking (query, NULL, &hls_seek_enabled, &hls_seek_start, &hls_seek_end);
+                if (hls_seek_enabled) {
+                    g_print ("Seeking is ENABLED from %" GST_TIME_FORMAT " to %" GST_TIME_FORMAT "\n",
+			     GST_TIME_ARGS (hls_seek_start), GST_TIME_ARGS (hls_seek_end));
+                } else {
+                    g_print ("Seeking is DISABLED for this stream.\n");
+                }
+            } else {
+                g_printerr ("Seeking query failed.");
+            }
+            gst_query_unref (query);
         }
         if (renderer_type[type]->autovideo) {
             char *sink = strstr(GST_MESSAGE_SRC_NAME(message), "-actual-sink-");
@@ -672,27 +849,50 @@ gboolean gstreamer_pipeline_bus_callback(GstBus *bus, GstMessage *message, void
     return TRUE;
 }
 
-void video_renderer_choose_codec (bool video_is_h265) {
+int video_renderer_choose_codec (bool video_is_h265) {
+    video_renderer_t *renderer_used = NULL;
+    video_renderer_t *renderer_unused = NULL;
     g_assert(!hls_video);
-    /* set renderer to h264 or h265, depending on pps/sps received by raop_rtp_mirror */
-    video_renderer_t *renderer_new = video_is_h265 ? renderer_type[1] : renderer_type[0];
-    if (renderer == renderer_new) {
-        return;
+    if (n_renderers == 1) {
+        if (video_is_h265) {
+            logger_log(logger, LOGGER_ERR, "video is h265 but the -h265 option was not used");
+            return -1;
+	}
+        renderer_used = renderer_type[0];
+    } else {
+        renderer_used = video_is_h265 ? renderer_type[1] : renderer_type[0];
+        renderer_unused = video_is_h265 ? renderer_type[0] : renderer_type[1];
     }
-    video_renderer_t *renderer_prev = renderer;
-    renderer = renderer_new;
+    if (renderer_used == NULL) { 
+        return -1;
+    } else if (renderer_used == renderer) {
+        return 0;
+    } else if (renderer) {
+        return -1;
+    }
+    renderer = renderer_used;
+    gst_element_set_state (renderer->pipeline, GST_STATE_PLAYING);
+    GstState old_state, new_state;
+    if (gst_element_get_state(renderer->pipeline, &old_state, &new_state, 100 * GST_MSECOND) == GST_STATE_CHANGE_FAILURE) {
+        g_error("video pipeline failed to go into playing state");
+	return -1;
+    }
+    logger_log(logger, LOGGER_DEBUG, "video_pipeline state change from %s to %s\n",
+               gst_element_state_get_name (old_state),gst_element_state_get_name (new_state));
     gst_video_pipeline_base_time = gst_element_get_base_time(renderer->appsrc);
-    /* it seems unlikely that the codec will change between h264 and h265 during a connection,
-     * but in case it does, we set the previous renderer to GST_STATE_NULL, detect
-     * when this is finished by listening for the bus message, and then reset it to
-     * GST_STATE_READY, so it can be reused if the codec changes again. */
-    if (renderer_prev) {
-        gst_app_src_end_of_stream (GST_APP_SRC(renderer_prev->appsrc));
-        gst_bus_set_flushing(renderer_prev->bus, TRUE);
-        /* set state of previous renderer to GST_STATE_NULL to (hopefully?) close its video window */
-        gst_element_set_state (renderer_prev->pipeline, GST_STATE_NULL);
-	renderer_prev->state_pending = true;     // will set state to PLAYING once state is NULL
+    if (renderer == renderer_type[1]) {
+        logger_log(logger, LOGGER_INFO, "*** video format is h265 high definition (HD/4K) video %dx%d", width, height);
     }
+    if (renderer_unused) {
+        for (int i = 0; i < n_renderers; i++) {
+            if (renderer_type[i] != renderer_unused) {
+                continue;
+            }
+            renderer_type[i] = NULL;
+            video_renderer_destroy_instance(renderer_unused);
+        }
+    }
+    return 0;
 }
 
 unsigned int video_reset_callback(void * loop) {
@@ -709,16 +909,18 @@ unsigned int video_reset_callback(void * loop) {
     return (unsigned  int) TRUE;
 }
 
-bool video_get_playback_info(double *duration, double *position, float *rate) {
+bool video_get_playback_info(double *duration, double *position, float *rate, bool *buffer_empty, bool *buffer_full) {
     gint64 pos = 0;
     GstState state;
     *duration = 0.0;
     *position = -1.0;
     *rate = 0.0f;
     if (!renderer) {
-   
         return true;
     }
+
+    *buffer_empty = (bool) hls_buffer_empty;
+    *buffer_full = (bool) hls_buffer_full;
     gst_element_get_state(renderer->pipeline, &state, NULL, 0);
     *rate = 0.0f;
     switch (state) {
@@ -728,12 +930,12 @@ bool video_get_playback_info(double *duration, double *position, float *rate) {
         break;
     }
 
-    if (!GST_CLOCK_TIME_IS_VALID(renderer->duration)) {
-        if (!gst_element_query_duration (renderer->pipeline, GST_FORMAT_TIME, &renderer->duration)) {
+    if (!GST_CLOCK_TIME_IS_VALID(hls_duration)) {
+        if (!gst_element_query_duration (renderer->pipeline, GST_FORMAT_TIME, &hls_duration)) {
             return true;
         }
     }
-    *duration = ((double) renderer->duration) / GST_SECOND;
+    *duration = ((double) hls_duration) / GST_SECOND;
     if (*duration) {
         if (gst_element_query_position (renderer->pipeline, GST_FORMAT_TIME, &pos) &&
                                         GST_CLOCK_TIME_IS_VALID(pos)) {
@@ -742,19 +944,27 @@ bool video_get_playback_info(double *duration, double *position, float *rate) {
     }
 
     logger_log(logger, LOGGER_DEBUG, "********* video_get_playback_info: position %" GST_TIME_FORMAT " duration %" GST_TIME_FORMAT " %s *********",
-               GST_TIME_ARGS (pos), GST_TIME_ARGS (renderer->duration), gst_element_state_get_name(state));
+               GST_TIME_ARGS (pos), GST_TIME_ARGS (hls_duration), gst_element_state_get_name(state));
 
     return true;
 }
 
+void video_renderer_set_start(float position) {
+    int pos_in_micros = (int) (position * SECOND_IN_MICROSECS);
+    hls_requested_start_position = (gint64) (pos_in_micros * GST_USECOND);
+    logger_log(logger, LOGGER_DEBUG, "register HLS video start position %f %lld", position,
+               hls_requested_start_position);    
+}
+
 void video_renderer_seek(float position) {
-    double pos = (double) position;
-    pos *=  GST_SECOND;
-    gint64 seek_position = (gint64) pos;
+    int pos_in_micros = (int) (position * SECOND_IN_MICROSECS);
+    gint64 seek_position = (gint64) (pos_in_micros * GST_USECOND);
+    /* don't seek to within 1  microsecond  of beginning or end of video */
+    if (hls_duration < 2000) return;
     seek_position =  seek_position < 1000 ? 1000 : seek_position;
-    seek_position =  seek_position > renderer->duration  - 1000 ? renderer->duration - 1000: seek_position;
+    seek_position =  seek_position > hls_duration  - 1000 ? hls_duration - 1000 : seek_position;
     g_print("SCRUB: seek to %f secs =  %" GST_TIME_FORMAT ", duration = %" GST_TIME_FORMAT "\n", position,
-            GST_TIME_ARGS(seek_position),  GST_TIME_ARGS(renderer->duration));
+            GST_TIME_ARGS(seek_position),  GST_TIME_ARGS(hls_duration));
     gboolean result = gst_element_seek_simple(renderer->pipeline, GST_FORMAT_TIME,
                                               (GstSeekFlags)(GST_SEEK_FLAG_FLUSH | GST_SEEK_FLAG_KEY_UNIT),
                                               seek_position);
